----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    13:53:06 04/29/2014 
-- Design Name: 
-- Module Name:    AMC_DATA_FIFO - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.std_logic_misc.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;
Library UNIMACRO;
use UNIMACRO.vcomponents.all;

entity AMC_DATA_FIFO is
    Port ( wclk : in  STD_LOGIC;
           rclk : in  STD_LOGIC;
           reset : in  STD_LOGIC;
           fifo_en : in  STD_LOGIC;
           we : in  STD_LOGIC;
           re : in  STD_LOGIC;
           Di : in  STD_LOGIC_VECTOR (63 downto 0);
           Do : out  STD_LOGIC_VECTOR (63 downto 0);
           full : out  STD_LOGIC
          );
end AMC_DATA_FIFO;

architecture Behavioral of AMC_DATA_FIFO is
type array15x12 is array(0 to 14) of std_logic_vector(11 downto 0);
signal RDCOUNT : array15x12;
signal WRCOUNT : array15x12;
--signal RDADDR : std_logic_vector(13 downto 0) := (others => '0');
--signal WRADDR : std_logic_vector(13 downto 0) := (others => '0');
--signal WRADDRP : std_logic_vector(12 downto 0) := (others => '0');
--signal WRADDR0SyncRegs : std_logic_vector(2 downto 0) := (others => '0');
--signal WRADDR1SyncRegs : std_logic_vector(2 downto 0) := (others => '0');
--signal WRADDR2SyncRegs : std_logic_vector(2 downto 0) := (others => '0');
signal FIFO_full : std_logic_vector(13 downto 0) := (others => '0');
signal FIFO_empty : std_logic_vector(13 downto 0) := (others => '0');
signal FIFO_RdEn : std_logic_vector(6 downto 0) := (others => '0');
signal FIFO_Do : std_logic_vector(62 downto 0) := (others => '0');
--signal RDEN : std_logic := '0';
--signal REGCE : std_logic := '0';
--signal Do63_validp : std_logic := '0';
--signal Do63_valid : std_logic := '0';
begin
full <= FIFO_full(0);
g_FIFO: for i in 0 to 6 generate
	i_FIFO_i : FIFO_DUALCLOCK_MACRO
		 generic map (
				DEVICE => "7SERIES",            -- Target Device: "VIRTEX5", "VIRTEX6", "7SERIES" 
				ALMOST_FULL_OFFSET => X"0004",  -- Sets almost full threshold
				ALMOST_EMPTY_OFFSET => X"0080", -- Sets the almost empty threshold
				DATA_WIDTH => 9,   -- Valid values are 1-72 (37-72 only valid when FIFO_SIZE="36Kb")
				FIFO_SIZE => "36Kb",            -- Target BRAM, "18Kb" or "36Kb" 
				FIRST_WORD_FALL_THROUGH => TRUE) -- Sets the FIFO FWFT to TRUE or FALSE
		 port map (
				ALMOSTEMPTY => open,   -- 1-bit output almost empty
				ALMOSTFULL => FIFO_full(i),     -- 1-bit output almost full
				DO => FIFO_Do(i*9+8 downto i*9),                     -- Output data, width defined by DATA_WIDTH parameter
				EMPTY => FIFO_empty(i),               -- 1-bit output empty
				FULL => open,                 -- 1-bit output full
				RDCOUNT => RDCOUNT(i),           -- Output read count, width determined by FIFO depth
				RDERR => open,               -- 1-bit output read error
				WRCOUNT => WRCOUNT(i),           -- Output write count, width determined by FIFO depth
				WRERR => open,               -- 1-bit output write error
				DI => Di(i*9+8 downto i*9),                     -- Input data, width defined by DATA_WIDTH parameter
				RDCLK => wclk,               -- 1-bit input read clock
				RDEN => FIFO_Rden(i),                 -- 1-bit input read enable
				RST => reset,                   -- 1-bit input reset
				WRCLK => wclk,               -- 1-bit input write clock
				WREN => we                  -- 1-bit input write enable
		 );
	i_FIFO_o : FIFO_DUALCLOCK_MACRO
		 generic map (
				DEVICE => "7SERIES",            -- Target Device: "VIRTEX5", "VIRTEX6", "7SERIES" 
				ALMOST_FULL_OFFSET => X"0004",  -- Sets almost full threshold
				ALMOST_EMPTY_OFFSET => X"0006", -- Sets the almost empty threshold
				DATA_WIDTH => 9,   -- Valid values are 1-72 (37-72 only valid when FIFO_SIZE="36Kb")
				FIFO_SIZE => "36Kb",            -- Target BRAM, "18Kb" or "36Kb" 
				FIRST_WORD_FALL_THROUGH => TRUE) -- Sets the FIFO FWFT to TRUE or FALSE
		 port map (
				ALMOSTEMPTY => FIFO_empty(i+7),   -- 1-bit output almost empty
				ALMOSTFULL => open,     -- 1-bit output almost full
				DO => Do(i*9+8 downto i*9),                     -- Output data, width defined by DATA_WIDTH parameter
				EMPTY => open,               -- 1-bit output empty
				FULL => FIFO_full(i+7),                 -- 1-bit output full
				RDCOUNT => RDCOUNT(i+7),           -- Output read count, width determined by FIFO depth
				RDERR => open,               -- 1-bit output read error
				WRCOUNT => WRCOUNT(i+7),           -- Output write count, width determined by FIFO depth
				WRERR => open,               -- 1-bit output write error
				DI => FIFO_Do(i*9+8 downto i*9),                     -- Input data, width defined by DATA_WIDTH parameter
				RDCLK => rclk,               -- 1-bit input read clock
				RDEN => re,                 -- 1-bit input read enable
				RST => reset,                   -- 1-bit input reset
				WRCLK => wclk,               -- 1-bit input write clock
				WREN => FIFO_Rden(i)                  -- 1-bit input write enable
		 );
	FIFO_Rden(i) <= fifo_en and not FIFO_full(i+7) and not FIFO_empty(i);
end generate;
i_FIFO63 : FIFO_DUALCLOCK_MACRO
   generic map (
      DEVICE => "7SERIES",            -- Target Device: "VIRTEX5", "VIRTEX6", "7SERIES" 
      ALMOST_FULL_OFFSET => X"0004",  -- Sets almost full threshold
      ALMOST_EMPTY_OFFSET => X"0006", -- Sets the almost empty threshold
      DATA_WIDTH => 1,   -- Valid values are 1-72 (37-72 only valid when FIFO_SIZE="36Kb")
      FIFO_SIZE => "18Kb",            -- Target BRAM, "18Kb" or "36Kb" 
      FIRST_WORD_FALL_THROUGH => TRUE) -- Sets the FIFO FWFT to TRUE or FALSE
   port map (
      ALMOSTEMPTY => open,   -- 1-bit output almost empty
      ALMOSTFULL => open,     -- 1-bit output almost full
      DO => Do(63 downto 63),                     -- Output data, width defined by DATA_WIDTH parameter
      EMPTY => open,               -- 1-bit output empty
      FULL => open,                 -- 1-bit output full
      RDCOUNT => RDCOUNT(14),           -- Output read count, width determined by FIFO depth
      RDERR => open,               -- 1-bit output read error
      WRCOUNT => WRCOUNT(14),           -- Output write count, width determined by FIFO depth
      WRERR => open,               -- 1-bit output write error
      DI => Di(63 downto 63),                     -- Input data, width defined by DATA_WIDTH parameter
      RDCLK => rclk,               -- 1-bit input read clock
      RDEN => re,                 -- 1-bit input read enable
      RST => reset,                   -- 1-bit input reset
      WRCLK => wclk,               -- 1-bit input write clock
      WREN => we                  -- 1-bit input write enable
   );
--i_FIFO63 : BRAM_SDP_MACRO
--   generic map (
--      BRAM_SIZE => "18Kb", -- Target BRAM, "18Kb" or "36Kb" 
--      DEVICE => "7SERIES", -- Target device: "VIRTEX5", "VIRTEX6", "7SERIES", "SPARTAN6" 
--      WRITE_WIDTH => 1,    -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
--      READ_WIDTH => 1,     -- Valid values are 1-72 (37-72 only valid when BRAM_SIZE="36Kb")
--      DO_REG => 1) -- Optional output register (0 or 1)
--   port map (
--      DO => Do(63 downto 63),         -- Output read data port, width defined by READ_WIDTH parameter
--      DI => Di(63 downto 63),         -- Input write data port, width defined by WRITE_WIDTH parameter
--      RDADDR => RDADDR, -- Input read address, width defined by read port depth
--      RDCLK => rclk,   -- 1-bit input read clock
--      RDEN => RDEN,     -- 1-bit input read port enable
--      REGCE => REGCE,   -- 1-bit input read output register enable
--      RST => '0',       -- 1-bit input reset 
--      WE => "1",         -- Input write enable, width defined by write port depth
--      WRADDR => WRADDR, -- Input write address, width defined by write port depth
--      WRCLK => wclk,   -- 1-bit input write clock
--      WREN => we      -- 1-bit input write port enable
--   );
--REGCE <= not Do63_valid or re;
--RDEN <= not Do63_validp or not Do63_valid or re; 
--process(wclk,reset)
--begin
--	if(reset = '1')then
--		WRADDR <= (others => '0');
--	elsif(wclk'event and wclk = '1')then
--		if(we = '1')then
--			case WRADDR(2 downto 0) is
--				when "000" => WRADDR(2 downto 0) <= "001";
--				when "001" => WRADDR(2 downto 0) <= "011";
--				when "011" => WRADDR(2 downto 0) <= "010";
--				when "010" => WRADDR(2 downto 0) <= "110";
--				when "110" => WRADDR(2 downto 0) <= "111";
--				when "111" => WRADDR(2 downto 0) <= "101";
--				when "101" => WRADDR(2 downto 0) <= "100";
--				when others => WRADDR(2 downto 0) <= "000";
--			end case;
--			if(WRADDR(2 downto 0) = "100")then
--				WRADDR(12 downto 3) <= WRADDR(12 downto 3) + 1;
--			end if;
--		end if;
--	end if;
--end process;
--WRADDRP(2) <= WRADDR2SyncRegs(2);
--WRADDRP(1) <= WRADDR1SyncRegs(2);
--WRADDRP(0) <= WRADDR0SyncRegs(2);
--process(rclk,reset)
--begin
--	if(reset = '1')then
--		WRADDR0SyncRegs <= (others => '0');
--		WRADDR1SyncRegs <= (others => '0');
--		WRADDR2SyncRegs <= (others => '0');
--		Do63_validp <= '0';
--		Do63_valid <= '0';
--		RDADDR <= (others => '0');
--		WRADDRP(12 downto 3) <= (others => '0');
--	elsif(rclk'event and rclk = '1')then
--		WRADDR0SyncRegs <= WRADDR0SyncRegs(1 downto 0) & WRADDR(0);
--		WRADDR1SyncRegs <= WRADDR1SyncRegs(1 downto 0) & WRADDR(1);
--		WRADDR2SyncRegs <= WRADDR2SyncRegs(1 downto 0) & WRADDR(2);
--		if(WRADDR2SyncRegs(2) = '1' and WRADDR2SyncRegs(1) = '0')then
--		  WRADDRP(12 downto 3) <= WRADDRP(12 downto 3) + 1;
--		end if;
--		if(WRADDRP /= RDADDR)then
--			Do63_validp <= '1';
--		elsif(re = '1' or Do63_valid = '0')then
--			Do63_validp <= '0';
--		end if;
--		if(Do63_validp = '1')then
--			Do63_valid <= '1';
--		elsif(re = '1')then
--			Do63_valid <= '0';
--		end if;
--		if(RDEN = '1' and WRADDRP /= RDADDR)then
--			case RDADDR(2 downto 0) is
--				when "000" => RDADDR(2 downto 0) <= "001";
--				when "001" => RDADDR(2 downto 0) <= "011";
--				when "011" => RDADDR(2 downto 0) <= "010";
--				when "010" => RDADDR(2 downto 0) <= "110";
--				when "110" => RDADDR(2 downto 0) <= "111";
--				when "111" => RDADDR(2 downto 0) <= "101";
--				when "101" => RDADDR(2 downto 0) <= "100";
--				when others => RDADDR(2 downto 0) <= "000";
--			end case;
--			if(RDADDR(2 downto 0) = "100")then
--				RDADDR(12 downto 3) <= RDADDR(12 downto 3) + 1;
--			end if;
--		end if;
--	end if;
--end process;
end Behavioral;

